import { isAbsolute, join, normalize, dirname, basename, extname, relative } from 'path';
import { createFilter } from 'rollup-pluginutils';
import fs, { readFile as readFile$1 } from 'fs';
import { promisify } from 'util';
import _glob from 'glob';
import acorn from 'acorn';

/*!
 * is-var-name | ISC (c) Shinnosuke Watanabe
 * https://github.com/shinnn/is-var-name
*/
function isVarName(str) {
	if (typeof str !== 'string') {
		return false;
	}

	if (str.trim() !== str) {
		return false;
	}

	try {
		new Function(str, 'var ' + str);
	} catch (e) {
		return false;
	}

	return true;
}

function setByPath(obj, val, path) {
	function _set(o, ks) {
		if (ks.length > 1) {
			let k = ks.shift();

			o[k] || (o[k] = {});

			_set(o[k], ks);
		} else {
			o[last(ks)] = val;
		}
	}

	_set(obj, path.split(/\//g));
}

const tryResolve = (target = "", exts = []) => {
	for (const ext of exts) {
		const fullfp = target + "." + ext;

		if (isExists(fullfp)) {
			return fullfp;
		}
	}

	return null;
};

const isDir = (fp) => fs.statSync(fp).isDirectory();

const isExists = (fp) => fs.existsSync(fp);

const validIdent = (ident) => isVarName(ident);

const last = (arraylike) => arraylike[arraylike.length - 1];

const dualEach = (o) => (fn) =>
	Object.entries(o).map(async ([k, v]) => await fn(k, v));

const all$p = (...ps) => Promise.all(ps);

const dualAll = (o) => (fn) => all$p(...dualEach(o)(fn));

const ERR_EXPORT_CONFLICT = "EXPORT_CONFLICT";
const ERR_INTERGRATION_TARGET = "INTERGRATION_TARGET";
const ERR_VARIABLE_MISSING = "VARIABLE_MISSING";
const ERR_OPTION_INVALID = "OPTION_INVALID";

class ResolveError extends Error {
	constructor(code, message) {
		super(`@zrlps/resolve: ${code}: ${message}`);
		this.code = code;
	}
}

function checkOptions({
	base = "",
	candidateExt = [],
	variables = {},
	dirBehaviour = "es6",
	include,
	exclude,
}) {
	const cwd = process.cwd();

	if (!isAbsolute(base)) {
		base = join(cwd, base);
	}

	if (!isExists(base)) {
		throw new ResolveError(
			ERR_OPTION_INVALID,
			`The option "base" (${base}) is not resolved.`
		);
	} else if (!isDir(base)) {
		throw new ResolveError(
			ERR_OPTION_INVALID,
			`The option "base" (${base}) must be a directory.`
		);
	}

	if (
		candidateExt.constructor !== Array ||
		!candidateExt.reduce((ok, v) => typeof v === "string" && ok, true)
	) {
		throw new ResolveError(
			ERR_OPTION_INVALID,
			`The option "candidateExt" must be an array of string. Received ${candidateExt}`
		);
	}

	if (variables.constructor !== Object) {
		throw new ResolveError(
			ERR_OPTION_INVALID,
			`The option "variables" must be an Object. Received ${variables}`
		);
	}

	if (!["es6", "collective", "auto"].includes(dirBehaviour)) {
		throw new ResolveError(
			ERR_OPTION_INVALID,
			`The option "dirBehaviour" must be "es6", "collective" or "auto". Received ${dirBehaviour}`
		);
	}

	variables = Object.entries(variables)
		.map(([name, dir]) => {
			let p = dir;

			if (!isAbsolute(dir)) {
				p = join(cwd, dir);
			}

			if (!isExists(p)) {
				throw new ResolveError(
					ERR_OPTION_INVALID,
					`The path "${dir}" in option "variables" is not resolved.`
				);
			}

			return [name, p];
		})
		.reduce((m, [name, p]) => m.set(name, p), new Map());

	return { base, variables, candidateExt, dirBehaviour, include, exclude };
}

var name = "@zrlps/rollup-plugin-resolve";

var preConfig = {
	sapper: {
		exclude: [/(svelte\/internal)/, /\.\/(internal|shared)/],
		base: "src",
		candidateExt: ["svelte"],
	},
};

const readFile = promisify(readFile$1);

function genExportAnalyzer() {
	// filepath str: { isDefault exports }
	// The filepath is absolute or relative with cwd
	const cache_exports = new Map();

	async function getExports(fp) {
		if (cache_exports.has(fp)) {
			return cache_exports.get(fp);
		}

		const code = await readFile(fp, "utf8");

		const option = { sourceType: "module" };
		const exports = {};
		let isDefault = false;

		for (const node of acorn.parse(code, option).body) {
			const { type, source } = node;

			if (type === "ExportAllDeclaration") {
				exports["*"] = source.value;
			} else if (type === "ExportDefaultDeclaration") {
				isDefault = true;
			} else if (type === "ExportNamedDeclaration") {
				if (node.declaration) {
					for (const declaration of node.declaration.declarations) {
						let { name } = declaration.id;
						exports[name] = "";
					}
				} else {
					for (const specifier of node.specifiers) {
						let { name } = specifier.exported;
						exports[name] = source ? source.value : "";
					}
				}
			}
		}

		const exportAnalysisResult = { isDefault, exports };

		cache_exports.set(fp, exportAnalysisResult);

		return exportAnalysisResult;
	}

	return {
		getExports,
		clear() {
			cache_exports.clear();
		},
	};
}

const glob = promisify(_glob);

const entries = Object.entries;
const validKeyPath = (kp) =>
	kp.split("/").reduce((t, k) => validIdent(k) && t, true);

function genVirtuaGrouplModuleMaker() {
	// filepath: VirtualCombineModule
	// The filepath is absolute or relative with cwd.
	const cache_combine = new Map();
	// filepath: VirtualIntergrationModule
	// The filepath's type is same with above.
	const cache_intergration = new Map();

	const { getExports, clear: exportAnalayzer_clear } = genExportAnalyzer();

	async function getVirtualModule(fp, isIntergration) {
		if (!isDir(fp)) return null;

		let module = isIntergration
			? cache_intergration.get(fp)
			: cache_combine.get(fp);

		if (module) return module;

		const m_empty = new Set(); // Set<ID>
		const m_all = new Set(); // Set<StylizedPath>
		const m_f_ids = {
			/* { StylizedPath: []ID } */
		};
		const m_defaults = {
			/* ID: StylizedPath */
		};
		const m_id_name = {
			/* { ID: name } */
		};
		const m_kp_id = {
			/* KeyPath: ID */
		};

		const genID = ((count = 0) => () => `_${count++}`)();

		const _filepaths = await glob(join(fp, isIntergration ? "/**" : "/*"));

		const filepaths = _filepaths.reduce((fps, f) => {
			for (const [fa] of fps.entries()) {
				if (fa === f) continue;
				fa.startsWith(f) && fps.delete(f);
			}
			return fps;
		}, new Set(_filepaths));

		const kpExported = new Set();

		for (let f of filepaths) {
			if (normalize(f) === normalize(fp)) continue;

			const dir_name = dirname(f);
			const filename = basename(f, extname(f));
			const is_dir = isDir(f);
			const kp = relative(
				fp,
				is_dir ? f : join(dir_name, filename)
			).replace("\\", "/");

			if (!is_dir) {
				const { isDefault, exports } = await getExports(f);

				if (!isIntergration) {
					m_all.add(f);

					await dualAll(exports)((name, source) => {
						const kp = relative(fp, join(dir_name, name));

						if (kpExported.has(kp)) {
							// TODO: need more information for error
							throw new ResolveError(
								ERR_EXPORT_CONFLICT,
								`Export conflict.`
							);
						}

						kpExported.add(kp);

						// named export has a higher priority
						let id = m_kp_id[kp];
						if (m_kp_id[kp]) {
							m_empty.delete(id);
							delete m_defaults[id];
						} else {
							id = genID();
						}

						if (name !== "*") {
							m_id_name[id] = kp;
							m_kp_id[kp] = id;
							(m_f_ids[f] = m_f_ids[f] || new Set()).add(id);
						}
					});

					if (validKeyPath(kp) && !m_kp_id[kp]) {
						// The default export has a lower priority.
						const id = genID();

						m_id_name[id] = kp;
						m_kp_id[kp] = id;
						if (isDefault) {
							m_defaults[id] = f;
						} else {
							m_empty.add(id);
						}
					}
				} else {
					if (validKeyPath(kp)) {
						// TODO:
						const id = genID();
						m_id_name[id] = kp;
						m_kp_id[kp] = id;

						if (isDefault) {
							m_defaults[id] = f;
						} else {
							m_empty.add(id);
						}
					}
				}
			} else {
				if (isIntergration && validKeyPath(kp)) {
					// TODO:
					const id = genID();
					m_kp_id[kp] = id;
					m_empty.add(id);
				}
			}
		}

		// Generate the code of virtual-module.
		const code = await genCode(
			{
				m_empty,
				m_all,
				m_f_ids,
				m_defaults,
				m_id_name,
				m_kp_id,
			},
			!isIntergration
		);

		module = { code };

		isIntergration
			? cache_intergration.set(fp, module)
			: cache_combine.set(fp, module);

		return module;
	}

	return {
		getVirtualModule,
		clear() {
			cache_combine.clear();
			cache_intergration.clear();
			exportAnalayzer_clear();
		},
	};
}

async function genCode(
	{ m_f_ids, m_defaults, m_id_name, m_empty, m_all, m_kp_id },
	isMinor
) {
	// Generate the code of virtual-module.

	const import_named_frag = (
		await dualAll(m_f_ids)((f, ids) => {
			const transform = (ids) =>
				ids.map((id) => `${m_id_name[id]} as ${id}`).join(", ");
			return `import { ${transform(Array.from(ids))} } from "${f}";`;
		})
	).join("\n");

	const import_default_frag = (
		await dualAll(m_defaults)((id, f) => `import ${id} from "${f}";`)
	).join("\n");

	const declarate_frag =
		m_empty.size === 0
			? ""
			: "const $;".replace(
					"$",
					Array.from(m_empty)
						.map((id) => `${id} = {}`)
						.join(", ")
			  );

	const export_all_frag = !m_all
		? ""
		: Array.from(m_all)
				.map((f) => `export * from "${f}";`)
				.join("\n");

	const export_default_frag = !m_kp_id
		? ""
		: "export default $;".replace(
				"$",
				JSON.stringify(
					entries(m_kp_id).reduce(
						(summary, [kp, id]) => (
							setByPath(summary, id, kp), summary
						),
						{}
					)
				).replace(/"([^"]+)":"([^"]+)"/g, `"$1": $2`)
		  );

	const export_named_frag = !isMinor
		? ""
		: (
				await dualAll(m_defaults)(
					(id) => `export { ${id} as ${m_id_name[id]} };`
				)
		  ).join("\n");

	return [
		import_named_frag,
		import_default_frag,
		declarate_frag,
		export_all_frag,
		export_named_frag,
		export_default_frag,
	]
		.filter(Boolean)
		.join("\n");
}

const maybeES6Export = (dirpath) =>
	tryResolve(join(dirpath, "index"), ["js", "node"]);

// more details from https://nodejs.org/api/modules.html#modules_file_modules
const extensions = ["mjs", "js", "json", "node"];

/*
	mode
		I Intergration
		C Combine
*/
const genVirtualID = (target, mode) => `\0${mode}::${target}`;
const parseVirtualID = (raw) => {
	if (raw.startsWith("\0")) {
		let ret = raw.slice(1).split("::");

		if (ret.length === 2) {
			return { mode: ret[0], target: ret[1] };
		}
	}

	return null;
};

const internalVariables = new Set(["~", "@"]);

var index = (options = {}) => {
	const {
		include,
		exclude,
		base,
		candidateExt,
		variables,
		dirBehaviour,
	} = checkOptions(options);
	const filter = createFilter(include, exclude);

	const { getVirtualModule, clear } = genVirtuaGrouplModuleMaker();

	return {
		name,
		async resolveId(importee, importer) {
			if (!filter(importee)) return null;

			let target = importee;
			let fullfp = "";
			let isIntergration = false;

			if (importee.startsWith("{") && importee.endsWith("}")) {
				target = importee.slice(1, -1);
				isIntergration = true;
			}

			const frags = target.split("/");
			const firstFrag = frags.shift();
			let prefix = "";

			if (internalVariables.has(firstFrag)) {
				if (firstFrag === "@") {
					prefix = base;
				} else if (firstFrag === "~") {
					prefix = process.env.HOME;
				}

				target = frags.join("/");
				fullfp = join(prefix, target);
			} else if (firstFrag.startsWith("$")) {
				// Variable import
				const variable = firstFrag.slice(1);

				if (variables.has(variable)) {
					prefix = variables.get(variable);
				} else {
					throw new ResolveError(
						ERR_VARIABLE_MISSING,
						`The variable "${variable}" is not defined.`
					);
				}

				target = frags.join("/");
				fullfp = join(prefix, target);
			} else if ([".", ".."].includes(firstFrag)) {
				fullfp = join(dirname(importer), target);
			} else {
				return null;
			}

			let is_exsit = isExists(fullfp);
			let is_dir = is_exsit && isDir(fullfp);

			if (is_exsit) {
				if (is_dir) {
					let mode = "I";

					if (!isIntergration) {
						const es6_fullfp = maybeES6Export(fullfp);
						if (
							dirBehaviour === "collective" ||
							(dirBehaviour === "auto" && !es6_fullfp)
						) {
							mode = "C";
						} else {
							// es6 import
							return es6_fullfp;
						}
					}

					return genVirtualID(fullfp, mode);
				} else if (isIntergration) {
					throw new ResolveError(
						ERR_INTERGRATION_TARGET,
						`Intergration import only works for directory.`
					);
				} else {
					return fullfp;
				}
			} else {
				const maybe = tryResolve(fullfp, [
					...extensions,
					...candidateExt,
				]);

				if (maybe && !isDir(maybe)) return maybe;

				return null;
			}
		},

		generateBundle() {
			clear();
		},

		async load(id) {
			const parsedVID = parseVirtualID(id);

			if (parsedVID) {
				const { mode, target } = parsedVID;

				return await getVirtualModule(target, mode === "I");
			}
		},
	};
};

export default index;
export { preConfig };
//# sourceMappingURL=rollup-plugin-resolve.es.js.map
